/*
 * Copyright (C) 2010 The Android Open Source Project
 * Copyright (C) 2011 Adam Nyb�ck
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package se.anyro.nfc_reader;

import java.nio.charset.Charset;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import se.anyro.nfc_reader.record.ParsedNdefRecord;
import android.app.Activity;
import android.app.AlertDialog;
import android.app.PendingIntent;
import android.content.Intent;
import android.nfc.NdefMessage;
import android.nfc.NdefRecord;
import android.nfc.NfcAdapter;
import android.nfc.Tag;
import android.nfc.tech.MifareClassic;
import android.nfc.tech.MifareUltralight;
import android.os.Bundle;
import android.os.Parcelable;
import android.view.LayoutInflater;
import android.widget.LinearLayout;
import android.widget.TextView;

/**
 * An {@link Activity} which handles a broadcast of a new tag that the device just discovered.
 */
public class TagViewer extends Activity 
{

	private static final SimpleDateFormat TIME_FORMAT = new SimpleDateFormat();
	private LinearLayout mTagContent;

	private NfcAdapter mAdapter;
	private PendingIntent mPendingIntent;
	private NdefMessage mNdefPushMessage;

	private AlertDialog mDialog;

	@Override
	protected void onCreate(Bundle savedInstanceState)
	{
		super.onCreate(savedInstanceState);

		setContentView(R.layout.tag_viewer);

		mTagContent = (LinearLayout) findViewById(R.id.list);

		resolveIntent(getIntent());

		mDialog = new AlertDialog.Builder(this).setNeutralButton("Ok", null).create();

		mAdapter = NfcAdapter.getDefaultAdapter(this);

		if (mAdapter == null) 
		{
			showMessage(R.string.error, R.string.no_nfc);
		}

		mPendingIntent = PendingIntent.getActivity(this, 0,
				new Intent(this, getClass()).addFlags(Intent.FLAG_ACTIVITY_SINGLE_TOP), 0);

		mNdefPushMessage = new NdefMessage(new NdefRecord[] { newTextRecord(
				"Message from NFC Reader :-)", Locale.ENGLISH, true) });
	}

	/**
	 * 
	 * @param title
	 * @param message
	 */
	private void showMessage(int title, int message) 
	{
		mDialog.setTitle(title);
		mDialog.setMessage(getText(message));
		mDialog.show();
	}


	/**
	 * 
	 * @param text
	 * @param locale
	 * @param encodeInUtf8
	 * @return
	 */
	private NdefRecord newTextRecord(String text, Locale locale, boolean encodeInUtf8)
	{
		byte[] langBytes = locale.getLanguage().getBytes(Charset.forName("US-ASCII"));

		Charset utfEncoding = encodeInUtf8 ? Charset.forName("UTF-8") : Charset.forName("UTF-16");
		byte[] textBytes = text.getBytes(utfEncoding);

		int utfBit = encodeInUtf8 ? 0 : (1 << 7);
		char status = (char) (utfBit + langBytes.length);

		byte[] data = new byte[1 + langBytes.length + textBytes.length];
		data[0] = (byte) status;
		System.arraycopy(langBytes, 0, data, 1, langBytes.length);
		System.arraycopy(textBytes, 0, data, 1 + langBytes.length, textBytes.length);

		return new NdefRecord(NdefRecord.TNF_WELL_KNOWN, NdefRecord.RTD_TEXT, new byte[0], data);
	}

	@Override
	protected void onResume() 
	{
		super.onResume();

		if (mAdapter != null) 
		{
			if (!mAdapter.isEnabled()) {
				showMessage(R.string.error, R.string.nfc_disabled);
			}
			mAdapter.enableForegroundDispatch(this, mPendingIntent, null, null);
			mAdapter.enableForegroundNdefPush(this, mNdefPushMessage);
		}
	}

	@Override
	protected void onPause() 
	{
		super.onPause();

		if (mAdapter != null)
		{
			mAdapter.disableForegroundDispatch(this);
			mAdapter.disableForegroundNdefPush(this);
		}
	}

	/**
	 * 
	 * @param intent
	 */
	private void resolveIntent(Intent intent)
	{
		String action = intent.getAction();

		if (NfcAdapter.ACTION_TAG_DISCOVERED.equals(action)
				|| NfcAdapter.ACTION_TECH_DISCOVERED.equals(action)
				|| NfcAdapter.ACTION_NDEF_DISCOVERED.equals(action)) 
		{
			Parcelable[] rawMsgs = intent.getParcelableArrayExtra(NfcAdapter.EXTRA_NDEF_MESSAGES);
			NdefMessage[] msgs;

			if (rawMsgs != null) 
			{
				msgs = new NdefMessage[rawMsgs.length];
				for (int i = 0; i < rawMsgs.length; i++)
				{
					msgs[i] = (NdefMessage) rawMsgs[i];
				}
			} 
			else 
			{
				// Unknown tag type
				byte[] empty = new byte[0];
				byte[] id = intent.getByteArrayExtra(NfcAdapter.EXTRA_ID);
				Parcelable tag = intent.getParcelableExtra(NfcAdapter.EXTRA_TAG);
				byte[] payload = dumpTagData(tag).getBytes();
				NdefRecord record = new NdefRecord(NdefRecord.TNF_UNKNOWN, empty, id, payload);
				NdefMessage msg = new NdefMessage(new NdefRecord[] { record });
				msgs = new NdefMessage[] { msg };
			}
			// Setup the views
			buildTagViews(msgs);
		}
	}

	/**
	 * 
	 * @param p
	 * @return
	 */
	private String dumpTagData(Parcelable p)
	{
		StringBuilder sb = new StringBuilder();
		Tag tag = (Tag) p;
		byte[] id = tag.getId();
		sb.append("Tag ID (hex): ").append(getHex(id)).append("\n");
		sb.append("Tag ID (dec): ").append(getDec(id)).append("\n");

		String prefix = "android.nfc.tech.";
		sb.append("Technologies: ");

		for (String tech : tag.getTechList()) 
		{
			sb.append(tech.substring(prefix.length()));
			sb.append(", ");
		}

		sb.delete(sb.length() - 2, sb.length());

		for (String tech : tag.getTechList()) 
		{
			if (tech.equals(MifareClassic.class.getName()))
			{
				sb.append('\n');
				MifareClassic mifareTag = MifareClassic.get(tag);
				String type = "Unknown";

				switch (mifareTag.getType()) 
				{
				case MifareClassic.TYPE_CLASSIC:
					type = "Classic";
					break;
				case MifareClassic.TYPE_PLUS:
					type = "Plus";
					break;
				case MifareClassic.TYPE_PRO:
					type = "Pro";
					break;
				}
				sb.append("Mifare Classic type: ");
				sb.append(type);
				sb.append('\n');

				sb.append("Mifare size: ");
				sb.append(mifareTag.getSize() + " bytes");
				sb.append('\n');

				sb.append("Mifare sectors: ");
				sb.append(mifareTag.getSectorCount());
				sb.append('\n');

				sb.append("Mifare blocks: ");
				sb.append(mifareTag.getBlockCount());
			}

			if (tech.equals(MifareUltralight.class.getName()))
			{
				sb.append('\n');

				MifareUltralight mifareUlTag = MifareUltralight.get(tag);

				String type = "Unknown";

				switch (mifareUlTag.getType())
				{
				case MifareUltralight.TYPE_ULTRALIGHT:
					type = "Ultralight";
					break;
				case MifareUltralight.TYPE_ULTRALIGHT_C:
					type = "Ultralight C";
					break;
				}
				sb.append("Mifare Ultralight type: ");
				sb.append(type);
			}
		}

		return sb.toString();
	}

	/**
	 * 
	 * @param bytes
	 * @return
	 */
	private String getHex(byte[] bytes) 
	{
		StringBuilder sb = new StringBuilder();

		for (int i = bytes.length - 1; i >= 0; --i)
		{
			int b = bytes[i] & 0xff;

			if (b < 0x10)
				sb.append('0');

			sb.append(Integer.toHexString(b));

			if (i > 0) {
				sb.append(" ");
			}
		}
		return sb.toString();
	}

	/**
	 * 
	 * @param bytes
	 * @return
	 */
	private long getDec(byte[] bytes) 
	{
		long result = 0;
		long factor = 1;

		for (int i = 0; i < bytes.length; ++i) 
		{
			long value = bytes[i] & 0xffl;
			result += value * factor;
			factor *= 256l;
		}
		return result;
	}

	/**
	 * 
	 * @param msgs
	 */
	void buildTagViews(NdefMessage[] msgs) 
	{
		if (msgs == null || msgs.length == 0)
		{
			return;
		}

		LayoutInflater inflater = LayoutInflater.from(this);
		LinearLayout content = mTagContent;

		// Parse the first message in the list
		// Build views for all of the sub records
		Date now = new Date();

		List<ParsedNdefRecord> records = NdefMessageParser.parse(msgs[0]);
		final int size = records.size();

		for (int i = 0; i < size; i++)
		{
			TextView timeView = new TextView(this);
			timeView.setText(TIME_FORMAT.format(now));
			content.addView(timeView, 0);
			ParsedNdefRecord record = records.get(i);
			content.addView(record.getView(this, inflater, content, i), 1 + i);
			content.addView(inflater.inflate(R.layout.tag_divider, content, false), 2 + i);
		}
	}

	@Override
	public void onNewIntent(Intent intent)
	{
		setIntent(intent);
		resolveIntent(intent);
	}
}
